from __future__ import unicode_literals

from typing import Any, Callable, Optional

from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import DynamicContainer, HSplit, ScrollbarMargin, Window
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.dimension import D, Dimension
from prompt_toolkit.mouse_events import MouseEvent
from prompt_toolkit.widgets import Frame

from app.context import ProcMuxContext
from app.log import logger
from app.tui.focus import FocusManager
from app.tui.keybindings import register_app_wide_configured_keybindings, register_configured_keybinding
from app.tui_state import FocusWidget


class SideBar:
    _right_padding = 5

    def __init__(
            self,
            focus_manager: FocusManager,
            on_start: Optional[Callable[[int], Any]] = None,
            on_stop: Optional[Callable[[int], Any]] = None,
            on_up: Optional[Callable[[int], Any]] = None,
            on_down: Optional[Callable[[int], Any]] = None,
            on_mouse_event: Optional[Callable[[MouseEvent], Any]] = None,
    ):
        self._focus_manager = focus_manager
        self._ctx = ProcMuxContext()
        self._on_start = on_start
        self._on_stop = on_stop
        self._on_up = on_up
        self._on_down = on_down
        self._on_mouse_event = on_mouse_event if on_mouse_event else lambda _me: NotImplemented
        self._filter_mode = False
        self._fixed_width = self._ctx.config.layout.processes_list_width
        self._cached_proc_name_to_filtered_idx = (
            '',
            {name: self._ctx.tui_state.get_process_index_by_name(name) for name in
             self._ctx.tui_state.process_name_list})
        self._filter_buffer = Buffer()
        self._list_control = FormattedTextControl(
            text=self._get_text_fragments,
            show_cursor=False,
            focusable=True,
            key_bindings=self._get_selection_key_bindings(),
        )
        self._buffer_control = BufferControl(
            buffer=self._filter_buffer,
            key_bindings=self._get_buffer_input_key_bindings())

        self._focus_manager.register_focusable_element(FocusWidget.SIDE_BAR_FILTER, self._buffer_control)
        self._focus_manager.register_focusable_element(FocusWidget.SIDE_BAR_SELECT, self._list_control)

        def get_filter_container():
            if self._filter_mode or self._filter_buffer.text:
                return Window(content=self._buffer_control,
                              height=D(min=1, max=1),
                              width=D(max=self._fixed_width - 1, weight=1))
            return Window(height=D(min=0, max=0),
                          width=D(max=self._fixed_width - 1, weight=1))

        self.container = Frame(
            title='Processes',
            style='class:sidebar',
            width=D(min=self._fixed_width, max=self._fixed_width, weight=1),
            body=HSplit([
                DynamicContainer(get_container=get_filter_container),
                Window(
                    content=self._list_control,
                    style="class:select-box",
                    height=Dimension(min=5),
                    cursorline=True,
                    right_margins=[ScrollbarMargin(display_arrows=True), ],
                )
            ]))

    def _get_filtered_process_name_list(self):
        if not self._filter_buffer.text:
            return self._ctx.tui_state.process_name_list
        prefix = self._ctx.config.layout.category_search_prefix

        def filter_against_category(search_text: str, proc_name: str) -> bool:
            search_text = search_text[len(prefix):]
            proc_entry = self._ctx.config.procs[proc_name]
            if not proc_entry.categories:
                return False
            categories = {c.lower() for c in proc_entry.categories}
            return search_text.lower() in categories

        def filter_against_name_and_meta(search_text: str, proc_name: str) -> bool:
            proc_entry = self._ctx.config.procs[proc_name]
            tags = set()
            if proc_entry.meta_tags:
                for tag in proc_entry.meta_tags:
                    tags.add(tag.lower())
            return search_text.lower() in proc_name or search_text.lower() in tags

        def filter_(search_text: str, proc_name: str) -> bool:
            if search_text.startswith(prefix):
                return filter_against_category(search_text, proc_name)
            return filter_against_name_and_meta(search_text, proc_name)

        return [name for name in self._ctx.tui_state.process_name_list if
                name and filter_(self._filter_buffer.text, name)]

    def get_filtered_index_for_process_name(self, proc_name) -> int:
        current_filter = ''
        if self._filter_buffer.text:
            current_filter = self._filter_buffer.text
        cache_key, proc_name_to_idx = self._cached_proc_name_to_filtered_idx
        if cache_key != current_filter:
            cache_key, proc_name_to_idx = (
                current_filter, {
                    name: idx for idx, name in enumerate(self._get_filtered_process_name_list())
                })
            self._cached_proc_name_to_filtered_idx = (cache_key, proc_name_to_idx)
        return proc_name_to_idx[proc_name]

    def _get_text_fragments(self):
        result = []
        for name in self._get_filtered_process_name_list():
            idx = self._ctx.tui_state.get_process_index_by_name(name)
            ctx = ProcMuxContext()
            manager = ctx.tui_state.terminal_managers[idx]
            status = "DOWN"
            status_fg = ctx.config.style.status_stopped_color
            if manager.is_running():
                status = "UP"
                status_fg = ctx.config.style.status_running_color

            target_width = self._fixed_width - self._right_padding - len(status)
            if len(name) > target_width:
                name = name[0:target_width - 3] + '...'
            name_fixed = f'%{target_width * -1}s' % name

            fg_color = ctx.config.style.unselected_process_color
            bg_color = ''
            pointer_char = " "
            if idx == self._ctx.tui_state.selected_process_idx:
                result.append(("[SetCursorPosition]", ""))
                bg_color = ctx.config.style.selected_process_bg_color
                fg_color = ctx.config.style.selected_process_color
                pointer_char = ctx.config.style.pointer_char
            result.append(( f'fg:{fg_color} bg:{bg_color} bold', f'{pointer_char}{name_fixed}', self._on_mouse_event))
            result.append(( f'fg:{status_fg} bg:{bg_color} bold', status, self._on_mouse_event))
            result.append(('', '\n', self._on_mouse_event))

        return result

    def _get_buffer_input_key_bindings(self):
        kb = KeyBindings()
        for keybinding in self._ctx.config.keybinding.filter + self._ctx.config.keybinding.submit_filter:
            @kb.add(keybinding)
            def _exit_filter(_event) -> None:
                logger.info('in _exit_filter')
                filtered_list = self._get_filtered_process_name_list()
                if filtered_list:
                    self._ctx.tui_state.selected_process_idx = \
                        self._ctx.tui_state.get_process_index_by_name(filtered_list[0])
                else:
                    self._ctx.tui_state.selected_process_idx = -1
                self._focus_manager.set_focus(self._list_control)
                self._filter_mode = False
        return kb

    def _get_selection_key_bindings(self):
        kb = KeyBindings()
        up = -1
        down = 1

        def move(direction: int) -> bool:
            tui_state = self._ctx.tui_state
            filtered_list = self._get_filtered_process_name_list()
            if not filtered_list:
                return False

            idx_in_main_list = self._ctx.tui_state.selected_process_idx
            if idx_in_main_list == -1:
                first_proc = filtered_list[0]
                tui_state.selected_process_idx = tui_state.get_process_index_by_name(first_proc)
                return True

            if len(filtered_list) < 2:
                return False

            name = self._ctx.tui_state.process_name_list[idx_in_main_list]
            idx_in_filtered_list = self.get_filtered_index_for_process_name(name)
            new_idx_in_filtered_list = ((idx_in_filtered_list + direction) % len(filtered_list))
            new_process_name = filtered_list[new_idx_in_filtered_list]
            self._ctx.tui_state.selected_process_idx = tui_state.get_process_index_by_name(new_process_name)
            return True

        def _go_up(_event) -> None:
            moved = move(up)
            if moved and self._on_up:
                self._on_up(self._ctx.tui_state.selected_process_idx)

        kb = register_configured_keybinding('up', _go_up, kb)

        def _go_down(_event) -> None:
            moved = move(down)
            if moved and self._on_down:
                self._on_down(self._ctx.tui_state.selected_process_idx)

        kb = register_configured_keybinding('down', _go_down, kb)

        def _start(_event) -> None:
            logger.info('in _start')
            if self._on_start:
                self._on_start(self._ctx.tui_state.selected_process_idx)

        kb = register_configured_keybinding('start', _start, kb)

        def _stop(_event) -> None:
            logger.info('in _stop')
            if self._on_stop:
                self._on_stop(self._ctx.tui_state.selected_process_idx)

        kb = register_configured_keybinding('stop', _stop, kb)

        def _quit(_event) -> None:
            logger.info('in _quit')
            application = get_app()

            def handle_process_done_to_quit():
                still_running = self._ctx.tui_state.has_running_processes
                logger.info(f'in handle_process_done_to_quit-  still running:  {still_running}')
                if not still_running:
                    application.exit()

            if self._ctx.tui_state.quitting:
                return  # avoid registering process done handler multiple times
            self._ctx.tui_state.quitting = True
            for tm in self._ctx.tui_state.terminal_managers:
                logger.info('_quit - registered process_done_handler')
                tm.register_process_done_handler(handle_process_done_to_quit)
            for tm in self._ctx.tui_state.terminal_managers:
                logger.info('_quit - sending kill signals')
                tm.send_kill_signal()
            if not self._ctx.tui_state.has_running_processes:
                application.exit()

        kb = register_configured_keybinding('quit', _quit, kb)

        def _filter(_event) -> None:
            logger.info('in _filter')
            self._filter_mode = True
            self._filter_buffer.text = ''
            app = get_app()
            app.invalidate()
            self._ctx.tui_state.selected_process_idx = -1
            self._focus_manager.set_focus(self._buffer_control)

        kb = register_configured_keybinding('filter', _filter, kb)

        def _view_docs(_event) -> None:
            logger.info('in _view_docs')
            self._focus_manager.toggle_docs_open()

        kb = register_configured_keybinding('docs', _view_docs, kb)
        kb = register_app_wide_configured_keybindings(self._focus_manager, kb)

        return kb

    def __pt_container__(self):
        return self.container



from __future__ import unicode_literals

from typing import Callable, List

from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import DynamicKeyBindings, KeyBindings
from prompt_toolkit.layout import ConditionalContainer, DynamicContainer, HSplit, Layout, VSplit
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.styles import Style
from ptterm import Terminal

from app.context import ProcMuxContext
from app.exec import TerminalManager
from app.tui.docs import DocsDialog
from app.tui.focus import FocusManager
from app.tui.help import HelpPanel
from app.tui.process_description import ProcessDescriptionPanel
from app.tui.side_bar import SideBar
from app.tui.style import height_100, width_100
from app.tui.terminal import TerminalPanel
from app.tui_state import FocusWidget


def _create_terminal(cmd: List[str],
                     before_exec: Callable,
                     on_done: Callable) -> Terminal:
    return Terminal(
        command=cmd,
        width=width_100,
        height=height_100,
        style='class:terminal',
        done_callback=on_done,
        before_exec_func=before_exec)


def _prep_tui_state():
    ctx = ProcMuxContext()
    if ctx.config.layout.sort_process_list_alpha:
        ctx.tui_state.process_name_list = sorted(list(ctx.config.procs.keys()))
    else:
        ctx.tui_state.process_name_list = list(ctx.config.procs.keys())
    ctx.tui_state.selected_process_name = None
    if ctx.config.procs:
        ctx.tui_state.selected_process_idx = 0
    terminal_managers = [
        TerminalManager(proc_name, _create_terminal)
        for proc_name in ctx.tui_state.process_name_list
    ]
    ctx.tui_state.terminal_managers = terminal_managers


def start_tui():
    _prep_tui_state()

    ctx = ProcMuxContext()

    def _on_terminal_change(term):
        terminal_wrapper.set_current_terminal(term)

    focus_manager = FocusManager(on_terminal_change=_on_terminal_change)
    terminal_wrapper = TerminalPanel(focus_manager=focus_manager)

    def _on_sidebar_mouse_event(mouse_event: MouseEvent):
        if mouse_event.event_type == MouseEventType.MOUSE_UP:
            _, y = mouse_event.position
            ctx.tui_state.selected_process_idx = y
            terminal_wrapper.set_terminal_terminal_by_process_idx(y)
            focus_manager.focus_to_sidebar()

    side_bar = SideBar(
        focus_manager=focus_manager,
        on_start=lambda proc_idx: terminal_wrapper.start_cmd_in_terminal(proc_idx),
        on_stop=lambda proc_idx: terminal_wrapper.stop_command(proc_idx),
        on_down=lambda proc_idx: terminal_wrapper.set_terminal_terminal_by_process_idx(proc_idx),
        on_up=lambda proc_idx: terminal_wrapper.set_terminal_terminal_by_process_idx(proc_idx),
        on_mouse_event=_on_sidebar_mouse_event
    )

    main_layout_container = HSplit([
        VSplit([
            side_bar,
            terminal_wrapper
        ]),
        ConditionalContainer(content=ProcessDescriptionPanel(),
                             filter=not ctx.config.layout.hide_process_description_panel),
        ConditionalContainer(content=HelpPanel(focus_manager=focus_manager),
                             filter=not ctx.config.layout.hide_help)
    ])
    docs_layout_container = HSplit([
        DocsDialog(focus_manager),
        ConditionalContainer(content=ProcessDescriptionPanel(),
                             filter=not ctx.config.layout.hide_process_description_panel),
        ConditionalContainer(content=HelpPanel(focus_manager=focus_manager),
                             filter=not ctx.config.layout.hide_help)
    ])

    def _get_layout_container():
        if ctx.tui_state.zoomed_in:
            return terminal_wrapper.get_current_terminal()
        elif ctx.tui_state.docs_open:
            return docs_layout_container
        return main_layout_container

    def _get_app_keybindings():
        if focus_manager.get_focused_widget() == FocusWidget.TERMINAL:
            return terminal_wrapper.get_keybindings()
        return KeyBindings()

    application = Application(
        layout=Layout(
            container=DynamicContainer(get_container=_get_layout_container),
            focused_element=side_bar),
        full_screen=True,
        mouse_support=ctx.config.enable_mouse,
        key_bindings=DynamicKeyBindings(get_key_bindings=_get_app_keybindings),
        style=Style(list((ctx.config.style.style_classes or {}).items())),
        color_depth=ctx.config.style.color_depth
    )

    def refresh_app():
        application.invalidate()

    for manager in ctx.tui_state.terminal_managers:
        manager.register_process_done_handler(refresh_app)
        manager.register_process_spawn_handler(refresh_app)
        manager.autostart_conditionally()
        refresh_app()

    application.run()



from app.config import ProcMuxConfig
from app.tui_state import TUIState


class NeverBootstrapped(Exception):
    pass


class ProcMuxContext:
    """
    ProxMuxContext is a singleton object that acts as a global handle to both
    shared TUI state and all parsed configuration points

    the ProxMuxContext must be bootstrapped before the TUI app is started
    """
    def __new__(cls):
        if not hasattr(cls, '_instance'):
            cls._instance = super(ProcMuxContext, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        if not self._instance:
            self._config = None
            self.tui_state = TUIState()

    def bootstrap(self, config: ProcMuxConfig):
        self._config = config
        self.tui_state = TUIState()

    def validate_init(self):
        if not self._config:
            raise NeverBootstrapped('bootstrap must be called before the processes service can be used')

    @property
    def config(self) -> ProcMuxConfig:
        self.validate_init()
        return self._config



from __future__ import unicode_literals

from prompt_toolkit.formatted_text import HTML, merge_formatted_text
from prompt_toolkit.layout import Window
from prompt_toolkit.layout.controls import FormattedTextControl

from app.context import ProcMuxContext


class ProcessDescriptionPanel:
    def __init__(self):
        self._ctx = ProcMuxContext()
        self.container = Window(
            height=1,
            content=FormattedTextControl(
                text=self._get_formatted_text,
                focusable=False,
                show_cursor=False
            ))

    def _get_formatted_text(self):
        idx = self._ctx.tui_state.selected_process_idx
        if idx < 0:
            return merge_formatted_text([HTML('')])
        name = self._ctx.tui_state.process_name_list[idx]
        desc = self._ctx.config.procs[name].description
        if not desc:
            desc = ''
        else:
            desc = " - " + desc
        result = [
            HTML(f'<b>{name}</b>{desc}')
        ]
        return merge_formatted_text(result)

    def __pt_container__(self):
        return self.container



from dataclasses import dataclass, field, fields
import os
from typing import Dict, List, Literal, Optional, OrderedDict, Union

import hiyapyco
from prompt_toolkit.output import ColorDepth


class MisconfigurationError(Exception):
    pass


@dataclass
class ProcessConfig:
    """
    shell:(str) command to run (exactly one of shell or cmd must be provided).
    cmd: (list) Array of command and args to run (exactly one of shell or cmd must be provided).
    cwd:(str) Set working directory for the process.
    env: (Dict[str,str]) - Set env variables. Object keys are variable names.
    add_path: string|array - Add entries to the PATH environment variable.
    autostart: bool - Start process when procmux starts.
    stop: "SIGINT"|"SIGTERM"|"SIGKILL" - default will SIGKILL
    """
    autostart: bool = False
    shell: Optional[str] = None
    cmd: Optional[List[str]] = None
    cwd: str = os.getcwd()
    stop: str = "SIGKILL"
    env: Dict[str, Optional[str]] = None
    add_path: Optional[Union[str, List[str]]] = None
    description: Optional[str] = None
    docs: Optional[str] = None
    categories: Optional[List[str]] = None
    meta_tags: Optional[List[str]] = None

    def __post_init__(self):
        self.validate()

    def validate(self):
        if not self.cmd and not self.shell:
            raise MisconfigurationError('shell or cmd is required for every proc definition')


@dataclass
class LayoutConfig:
    hide_help: bool = False
    hide_process_description_panel: bool = False
    processes_list_width: int = 31
    sort_process_list_alpha: bool = True
    category_search_prefix: str = 'cat:'


@dataclass
class StyleConfig:
    selected_process_color: str = 'ansiblack'
    selected_process_bg_color: str = 'ansimagenta'
    unselected_process_color: str = 'ansiblue'
    status_running_color: str = 'ansigreen'
    status_stopped_color: str = 'ansired'
    placeholder_terminal_bg_color: str = '#1a1b26'
    pointer_char: str = '▶'
    style_classes: Optional[Dict[str, str]] = None
    color_level: \
        Optional[
            Union[Literal['monochrome'], Literal['ansicolors'], Literal['256colors'], Literal['truecolors']]] = None

    @property
    def color_depth(
            self
    ) -> Union[Literal['DEPTH_1_BIT'], Literal['DEPTH_4_BIT'], Literal['DEPTH_8_BIT'], Literal['DEPTH_24_BIT']]:
        if self.color_level == 'monochrome':
            return ColorDepth.MONOCHROME
        if self.color_level == 'ansicolors':
            return ColorDepth.ANSI_COLORS_ONLY
        if self.color_level == '256colors':
            return ColorDepth.DEFAULT
        return ColorDepth.TRUE_COLOR


@dataclass
class KeybindingConfig:
    quit: List[str] = field(default_factory=lambda: ['q'])
    filter: List[str] = field(default_factory=lambda: ['/'])
    submit_filter: List[str] = field(default_factory=lambda: ['enter'])
    start: List[str] = field(default_factory=lambda: ['s'])
    stop: List[str] = field(default_factory=lambda: ['x'])
    up: List[str] = field(default_factory=lambda: ['up', 'k'])
    down: List[str] = field(default_factory=lambda: ['down', 'j'])
    switch_focus: List[str] = field(default_factory=lambda: ['c-w'])
    zoom: List[str] = field(default_factory=lambda: ['c-z'])
    docs: List[str] = field(default_factory=lambda: ['?'])
    toggle_scroll: List[str] = field(default_factory=lambda: ['c-s'])

    def __post_init__(self):
        for keybinding_field in fields(KeybindingConfig):
            keybinding = getattr(self, keybinding_field.name)
            if type(keybinding) == str:
                keybinding = [keybinding]
                setattr(self, keybinding_field.name, keybinding)


@dataclass
class ProcMuxConfig:
    procs: Optional[Dict[str, ProcessConfig]]
    style: StyleConfig = StyleConfig()
    keybinding: KeybindingConfig = KeybindingConfig()
    shell_cmd: List[str] = field(default_factory=lambda: [os.environ['SHELL'], '-c'])
    layout: LayoutConfig = LayoutConfig()
    log_file: Optional[str] = None
    enable_mouse: bool = True

    def __post_init__(self):
        def is_dict_like(obj):
            return type(obj) == dict or isinstance(obj, OrderedDict)

        if is_dict_like(self.procs):
            process_config_data = dict()
            for proc_key in self.procs:
                process_config_data[proc_key] = ProcessConfig(**self.procs[proc_key])
            self.procs = process_config_data
        if is_dict_like(self.style):
            self.style = StyleConfig(**self.style)
        if is_dict_like(self.keybinding):
            self.keybinding = KeybindingConfig(**self.keybinding)
        if is_dict_like(self.layout):
            self.layout = LayoutConfig(**self.layout)


def parse_config(config_file: str,
                 override_config_file: Optional[str] = None) -> ProcMuxConfig:
    config_files = [config_file]
    if override_config_file:
        config_files.append(override_config_file)

    config_dict = hiyapyco.load(*config_files, method=hiyapyco.METHOD_SIMPLE,
                                failonmissingfiles=True)
    return ProcMuxConfig(**config_dict)



from app.__main__ import run_app
from app.config import parse_config
from app.args import cli_args


def start_cli():
    config = parse_config(cli_args.config, cli_args.config_override)
    run_app(config)


if __name__ == '__main__':
    start_cli()



from dataclasses import dataclass, field
from enum import auto, Enum
from typing import Dict, List, Optional
from app.exec import TerminalManager


class FocusWidget(Enum):
    SIDE_BAR_FILTER = auto()
    SIDE_BAR_SELECT = auto()
    TERMINAL = auto()
    DOCS = auto()


@dataclass
class TUIState:
    _process_name_to_idx: Optional[Dict[str, int]] = None
    selected_process_idx: Optional[int] = None
    process_name_list: List[str] = field(default_factory=lambda: [])
    process_status_list: List[bool] = field(default_factory=lambda: [])
    terminal_managers: List[TerminalManager] = field(default_factory=lambda: [])
    zoomed_in: bool = False
    docs_open: bool = False
    quitting: bool = False

    @property
    def selected_process_terminal_manager(self) -> Optional[TerminalManager]:
        if self.selected_process_idx is not None and self.selected_process_idx > -1:
            return self.terminal_managers[self.selected_process_idx]
        return None

    @property
    def is_selected_process_running(self) -> bool:
        return self.selected_process_terminal_manager.is_running() \
            if self.selected_process_terminal_manager else False

    @property
    def selected_process_has_terminal(self) -> bool:
        return \
            self.selected_process_terminal_manager.get_terminal() is not None \
            if self.selected_process_terminal_manager else False

    @property
    def has_running_processes(self) -> bool:
        for tm in self.terminal_managers:
            if tm.is_running():
                return True
        return False

    def get_process_index_by_name(self, proc_name):
        if not self._process_name_to_idx:
            self._process_name_to_idx = {name: idx for idx, name in enumerate(self.process_name_list)}
        idx = self._process_name_to_idx[proc_name]
        return idx

    def get_terminal_manager_by_name(self, proc_name):
        idx = self.get_process_index_by_name(proc_name)
        return self.terminal_managers[idx]



from prompt_toolkit.layout import DynamicContainer, Window
from prompt_toolkit.widgets import Frame

from app.context import ProcMuxContext
from app.tui.focus import FocusManager
from app.tui.keybindings import register_app_wide_configured_keybindings
from app.tui.style import height_100, width_100


class TerminalPanel:
    def __init__(
            self,
            focus_manager: FocusManager,
    ):
        self._focus_manager = focus_manager
        self._ctx = ProcMuxContext()
        self._terminal_placeholder = Window(
            style=f'bg:{self._ctx.config.style.placeholder_terminal_bg_color}',
            width=width_100,
            height=height_100)
        self._current_terminal = self._terminal_placeholder

        self._container = Frame(title='Terminal',
                                body=DynamicContainer(get_container=lambda: self._current_terminal))

    def start_cmd_in_terminal(self, proc_idx: int):
        if self._ctx.tui_state.quitting:
            return  # if procmux is in the process of quitting, don't start a new process
        self._current_terminal = self._ctx.tui_state.terminal_managers[proc_idx].spawn_terminal()

    def stop_command(self, proc_idx: int):
        self._ctx.tui_state.terminal_managers[proc_idx].send_kill_signal()

    def set_current_terminal(self, term):
        self._current_terminal = term

    def get_current_terminal(self):
        self._current_terminal

    def set_terminal_terminal_by_process_idx(self, proc_idx: int):
        term = self._ctx.tui_state.terminal_managers[proc_idx].get_terminal()
        if not term:
            term = self._terminal_placeholder
        self._current_terminal = term

    def get_keybindings(self):
        kb = register_app_wide_configured_keybindings(self._focus_manager)
        return kb

    def __pt_container__(self):
        return self._container



import os
import signal
import sys
from typing import Callable, List, Optional

from ptterm import Terminal

from app.log import logger


class TerminalManager:
    def __init__(
            self,
            process_name: str,
            terminal_ctor: Callable[[List[str]], Terminal],
            on_done: Optional[Callable] = None,
            on_spawn: Optional[Callable] = None
    ):
        from app.context import ProcMuxContext
        self._ctx = ProcMuxContext()
        self._term_ctor = terminal_ctor
        self._proc_name = process_name
        self._terminal: Terminal = None
        self._running = False
        self._scroll_mode = False
        self._on_done_callbacks = []
        if on_done:
            self._on_done_callbacks = [on_done]

        self._on_spawn_callbacks = []
        if on_spawn:
            self._on_spawn_callbacks = [on_spawn]

    def _export_env_vars(self) -> bool:
        proc_config = self._ctx.config.procs[self._proc_name]
        if proc_config.env:
            for key, value in proc_config.env.items():
                os.environ[key] = value
            return True
        return False

    def _adjust_path(self):
        proc_config = self._ctx.config.procs[self._proc_name]
        if proc_config.add_path:
            paths = proc_config.add_path
            if type(proc_config.add_path) == str:
                paths = [proc_config.add_path]
            for p in paths:
                sys.path.append(p)

    def _change_working_directory(self):
        proc_config = self._ctx.config.procs[self._proc_name]
        if proc_config.cwd:
            os.chdir(proc_config.cwd)

    def get_cmd(self) -> List[str]:
        proc_config = self._ctx.config.procs[self._proc_name]
        shell_cmd = self._ctx.config.shell_cmd
        cmd = []
        if proc_config.shell:
            cmd.extend(shell_cmd)
            cmd.append(proc_config.shell)
        else:
            cmd = proc_config.cmd
        return cmd

    def is_running(self) -> bool:
        return self._terminal and self._running

    def send_kill_signal(self):
        if self.is_running():
            stop_signal = self._ctx.config.procs[self._proc_name].stop
            sig_code = getattr(signal, stop_signal)
            try:
                if hasattr(self._terminal.process.terminal, 'send_signal'):
                    logger.info(f'stopping process with defined signal {stop_signal}')
                    self._terminal.process.terminal.send_signal(sig_code)
                else:
                    logger.info(f'using x-platform process kill() - disregarding defined kill sig')
                    self.kill()
            except Exception as e:
                logger.error(f'failed to kill process name: {self._proc_name} {e}')

    def kill(self):
        if self.is_running():
            try:
                logger.info('running kill()')
                self._terminal.terminal_control.process.kill()
            except Exception as e:
                logger.error(f'failed to kill process name: {self._proc_name} {e}')

    def _handle_process_done(self):
        logger.info(f'{self._proc_name} is done')
        self._running = False
        for handler in self._on_done_callbacks:
            handler()

    def _handle_process_spawned(self):
        logger.info(f'created terminal {self._terminal}')
        self._running = True
        for handler in self._on_spawn_callbacks:
            handler()

    def spawn_terminal(self):
        if self.is_running():
            logger.info(f'spawned - returning existing terminal - {self._terminal}')
            return self._terminal

        def before_exec():
            self._change_working_directory()
            self._export_env_vars()
            self._adjust_path()

        self._terminal = self._term_ctor(self.get_cmd(), before_exec, self._handle_process_done)
        terminal_index = self._ctx.tui_state.get_process_index_by_name(self._proc_name)
        if terminal_index != self._ctx.tui_state.selected_process_idx:
            logger.info('manually starting process because this terminal is not actively selected')
            self._terminal.process.start()

        self._handle_process_spawned()
        return self._terminal

    def get_terminal(self):
        return self._terminal

    def register_process_done_handler(self, on_done: Callable):
        self._on_done_callbacks.append(on_done)

    def register_process_spawn_handler(self, on_spawn: Callable):
        self._on_spawn_callbacks.append(on_spawn)

    def autostart_conditionally(self) -> bool:
        proc_config = self._ctx.config.procs[self._proc_name]
        if proc_config.autostart:
            logger.info(f'autostarting {self._proc_name}')
            term = self.spawn_terminal()

            return True
        return False

    def toggle_scroll_mode(self) -> bool:
        if not self._terminal:
            return self._scroll_mode
        if self._scroll_mode:
            self._terminal.exit_copy_mode()
            self._scroll_mode = False
        else:
            self._terminal.enter_copy_mode()
            self._scroll_mode = True
        return self._scroll_mode



import os
import select
import tempfile

import pyte

from app.__main__ import run_app
from app.config import parse_config


def _get_config_yaml(log_file: str) -> str:
    config_yaml = f"""\
    layout:
      hide_help: false
    keybinding:
      switch_focus: 'w' 
      up: 'k' 
      down: 'j' 
    procs:
      "tail log":
        shell: "tail -f {log_file}"
        autostart: true 
        description: 'tail the app log'
        docs: long description for tail log
        meta_tags:
          - 'follow'
      "print envs":
        shell: "echo $SOME_TEST"
        description: 'this command will print env vars that are configured in the child pid'
        env: 
          SOME_TEST: "ENV_VARS_ARE_WORKING"
        categories:
          - 'echo'
      "vim":
        shell: "vim"
        autostart: false
        description: 'start vim'
      "test long running proc":
        shell: "echo 'some text here' && sleep 3 && echo 'still running'  && sleep 3 && echo 'final text'"
        autostart: false
        description: 'print a using sleeps in between'
        categories:
          - 'echo'
    """
    return config_yaml


def preform_test_within_tui(keys, assertion):
    """Test TUI."""
    # create pseudo-terminal
    pid, f_d = os.forkpty()
    if pid == 0:
        # child process spawns TUI
        with tempfile.NamedTemporaryFile() as yaml_tmp:
            with tempfile.NamedTemporaryFile() as log_tmp:
                config_yaml = _get_config_yaml(log_tmp.name)
                yaml_tmp.write(config_yaml.encode('utf8'))
                yaml_tmp.flush()
                config = parse_config(yaml_tmp.name)
                run_app(config)

    else:
        screen = pyte.Screen(80, 30)
        stream = pyte.ByteStream(screen)
        for key in keys:
            print(f'sending key input: {key}')
            os.write(f_d, str.encode(key))
        # scrape pseudo-terminal's screen
        while True:
            try:
                [f_d], _, _ = select.select(
                    [f_d], [], [], 1)
            except (KeyboardInterrupt, ValueError):
                # either test was interrupted or the
                # file descriptor of the child process
                # provides nothing to be read
                break
            else:
                try:
                    # scrape screen of child process
                    data = os.read(f_d, 1024)
                    stream.feed(data)
                except OSError:
                    # reading empty
                    break
        print('\n')
        for line in screen.display:
            print(line)

        assertion(screen)


def join_screen_to_str(screen) -> str:
    return "\n".join(screen.display)


def test_tui_shows_process_list_help():
    def assert_process_list_help_bar(screen):
        full_screen = join_screen_to_str(screen)
        assert '<s> start' in full_screen
        assert '<x> stop' not in full_screen
        assert '<q> quit' in full_screen

    preform_test_within_tui(keys=[], assertion=assert_process_list_help_bar)


def test_tui_shows_terminal_help():
    def assert_terminal_help_bar(screen):
        full_screen = join_screen_to_str(screen)
        assert '<s> start' not in full_screen
        assert '<x> stop' not in full_screen
        assert '<w> switch' in full_screen

    preform_test_within_tui(keys=['s', 'w'], assertion=assert_terminal_help_bar)


def test_tui_filter():
    def assert_filter(screen):
        full_screen = join_screen_to_str(screen)
        assert '<s> start' not in full_screen
        assert '<enter> filter' in full_screen
        assert 'test long running proc' in full_screen
        assert 'tail log' not in full_screen

    preform_test_within_tui(keys=['/', *list('test long run')], assertion=assert_filter)


def test_tui_env_vars_in_child_pid():
    def assert_env_vars_printed(screen):
        full_screen = join_screen_to_str(screen)
        assert 'ENV_VARS_ARE_WORKING' in full_screen

    preform_test_within_tui(keys=['/', *list('print env'), '/', 'j', 's'], assertion=assert_env_vars_printed)


def test_tui_filter_for_missing_process_shows_no_results():
    def assert_no_results(screen):
        full_screen = join_screen_to_str(screen)
        assert 'tail log' not in full_screen
        assert 'print envs' not in full_screen
        assert 'vim' not in full_screen
        assert 'test long running proc' not in full_screen

    preform_test_within_tui(keys=['/', *list('NEVER'), '/', 'j'], assertion=assert_no_results)

def test_tui_filter_against_category():
    def assert_no_results(screen):
        full_screen = join_screen_to_str(screen)
        assert 'test long running proc' in full_screen
        assert 'print envs' in full_screen
        assert 'vim' not in full_screen

    preform_test_within_tui(keys=['/', *list('cat:echo'), '/', 'j'], assertion=assert_no_results)
def test_tui_filter_against_meta_tags():
    def assert_no_results(screen):
        full_screen = join_screen_to_str(screen)
        assert 'test long running proc' not in full_screen
        assert 'print envs' not in full_screen
        assert 'vim' not in full_screen
        assert 'tail log' in full_screen

    preform_test_within_tui(keys=['/', *list('follow'), '/', 'j'], assertion=assert_no_results)
def test_long_docs_get_displayed():
    def assert_no_results(screen):
        full_screen = join_screen_to_str(screen)
        assert 'long description for tail log' in full_screen
        assert 'print envs' not in full_screen

    preform_test_within_tui(keys=['/', *list('tail log'), '/', 'j', '?'], assertion=assert_no_results)


def test_tui_autostart():
    def assert_autostart(screen):
        for line in screen.display:
            if "tail log" in line:
                assert "UP" in line
                break
        else:
            raise RuntimeError('Active tail log process not found')

    preform_test_within_tui(keys=[], assertion=assert_autostart)



from prompt_toolkit import HTML
from prompt_toolkit.formatted_text import merge_formatted_text
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import FormattedTextControl, Window
from prompt_toolkit.widgets import Frame

from app.context import ProcMuxContext
from app.log import logger
from app.tui_state import FocusWidget


class DocsDialog:
    def __init__(self, focus_manager):
        self._focus_manager = focus_manager

        self._ctx = ProcMuxContext()
        self._container = Frame(
            title=self._get_title,
            key_bindings=self._get_key_bindings(),
            body=Window(
                content=FormattedTextControl(
                    text=self._get_formatted_text,
                    focusable=True,
                    show_cursor=False
                )))
        self._focus_manager.register_focusable_element(FocusWidget.DOCS, self._container)

    def _get_key_bindings(self):
        kb = KeyBindings()

        for keybinding in self._ctx.config.keybinding.docs:
            @kb.add(keybinding)
            def _close_docs(_event) -> None:
                logger.info('in _close_docs')
                self._focus_manager.toggle_docs_open()
        return kb
    def _get_title(self):
        idx = self._ctx.tui_state.selected_process_idx
        if idx > -1:
            return self._ctx.tui_state.process_name_list[idx]
        return "Help"

    def _get_formatted_text(self):
        idx = self._ctx.tui_state.selected_process_idx
        if idx > -1:
            result = []
            proc_name = self._ctx.tui_state.process_name_list[idx]
            proc_config = self._ctx.config.procs[proc_name]
            if proc_config.description:
                result.append(HTML(f'<b>{proc_config.description}</b>\n'))
            if proc_config.docs:
                result.append(HTML(proc_config.docs))
            if len(result) == 0:
                result.append(f'No docs available for process: {proc_name}')

            return merge_formatted_text(result)
        return HTML("No process is currently selected.")

    def __pt_container__(self):
        return self._container



from __future__ import unicode_literals

from typing import List

from prompt_toolkit.formatted_text import HTML, merge_formatted_text
from prompt_toolkit.layout import Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.widgets import Frame

from app.context import ProcMuxContext
from app.tui.focus import FocusManager
from app.tui_state import FocusWidget


class HelpPanel:
    def __init__(
            self,
            focus_manager: FocusManager
    ):
        self._ctx = ProcMuxContext()
        self._focus_manager = focus_manager
        self.container = Window(
            height=1,
            content=FormattedTextControl(
                text=self._get_formatted_text,
                focusable=False,
                show_cursor=False
            ))

    def _get_formatted_text(self):
        result = []
        delimiter = " | "
        key_config = self._ctx.config.keybinding
        selected_process_running = self._ctx.tui_state.is_selected_process_running
        selected_process_has_terminal = self._ctx.tui_state.selected_process_has_terminal
        if self._focus_manager.get_focused_widget() == FocusWidget.SIDE_BAR_SELECT:
            result.append(self._get_key_combo_text(key_config.up, 'up'))
            result.append(delimiter)
            result.append(self._get_key_combo_text(key_config.down, 'down'))
            if not selected_process_running and not self._ctx.tui_state.quitting:
                result.append(delimiter)
                result.append(self._get_key_combo_text(key_config.start, 'start'))
            if selected_process_running:
                result.append(delimiter)
                result.append(self._get_key_combo_text(key_config.stop, 'stop'))
            result.append(delimiter)
            result.append(self._get_key_combo_text(key_config.quit, 'quit'))
            if selected_process_has_terminal:
                result.append(delimiter)
                result.append(self._get_key_combo_text(key_config.switch_focus, 'switch focus'))
                result.append(delimiter)
                result.append(self._get_key_combo_text(key_config.zoom, 'zoom'))
                result.append(delimiter)
                result.append(self._get_key_combo_text(key_config.toggle_scroll, 'toggle scroll'))
        elif self._focus_manager.get_focused_widget() == FocusWidget.SIDE_BAR_FILTER:
            result.append(self._get_key_combo_text(key_config.submit_filter, 'filter'))
        elif self._focus_manager.get_focused_widget() == FocusWidget.DOCS:
            result.append(self._get_key_combo_text(key_config.docs, 'docs'))
        else:
            result.append(self._get_key_combo_text(key_config.switch_focus, 'switch focus'))
            result.append(delimiter)
            result.append(self._get_key_combo_text(key_config.zoom, 'zoom'))
            result.append(delimiter)
            result.append(self._get_key_combo_text(key_config.toggle_scroll, 'toggle scroll'))
        return merge_formatted_text(result)

    def _get_key_combo_text(self, key_combos: List[str], label: str):
        [first, *_] = key_combos
        return HTML(f'<b>&lt;{first}&gt;</b> {label}')

    def __pt_container__(self):
        return self.container



from typing import Callable

from prompt_toolkit.application import get_app

from app.context import ProcMuxContext
from app.log import logger
from app.tui_state import FocusWidget


class FocusManager:
    def __init__(self,
                 on_terminal_change: Callable):
        self._ctx = ProcMuxContext()
        self._focus_elements = []
        self._on_terminal_change = on_terminal_change

    def is_focused_on_free_form_input(self):
        return self.get_focused_widget() == FocusWidget.SIDE_BAR_FILTER

    def register_focusable_element(self, widget_type: FocusWidget, element):
        self._focus_elements.append((widget_type, element))

    def get_focused_widget(self) -> FocusWidget:
        application = get_app()
        for widget_type, element in self._focus_elements:
            if application.layout.has_focus(element):
                return widget_type
        return FocusWidget.TERMINAL

    def get_element_by_focus_widget(self, focus_widget: FocusWidget):
        items = [el for w, el in self._focus_elements if w == focus_widget]
        if items:
            return items[0]
        return None

    def set_focus(self, element):
        application = get_app()
        application.layout.focus(element)

    def toggle_zoom(self):
        zoomed_in = self._ctx.tui_state.zoomed_in
        if zoomed_in:
            logger.info('setting zoomed_in to False')
            self._ctx.tui_state.zoomed_in = False
            self.focus_to_sidebar()
        else:
            success = self.focus_to_terminal()
            if success:
                logger.info('setting zoomed_in to True')
                self._ctx.tui_state.zoomed_in = True

    def toggle_docs_open(self):
        docs_open = self._ctx.tui_state.docs_open
        logger.info(f'setting docks_open to {not docs_open}')
        self._ctx.tui_state.docs_open = not docs_open
        if self._ctx.tui_state.docs_open:
            element = self.get_element_by_focus_widget(FocusWidget.DOCS)
            if element:
                self.set_focus(element)
        else:
            element = self.get_element_by_focus_widget(FocusWidget.SIDE_BAR_SELECT)
            if element:
                self.set_focus(element)

    def focus_to_terminal(self) -> bool:
        idx = self._ctx.tui_state.selected_process_idx
        if idx > -1:
            t_manager = self._ctx.tui_state.terminal_managers[idx]
            term = t_manager.get_terminal()
            if term:
                application = get_app()
                self._on_terminal_change(term)
                application.layout.focus(term)
                logger.info(f'focusing on terminal for idx: {idx} ')
                return True
        return False

    def toggle_sidebar_terminal_focus(self):
        assert self._on_terminal_change
        if self._ctx.tui_state.docs_open:
            logger.info('in toggle_sidebar_terminal_focus - but the docs are open, '
                        'toggling docs off instead')
            self.toggle_docs_open()
            return
        if self._ctx.tui_state.zoomed_in:
            logger.info('in toggle_sidebar_terminal_focus - currently zoomed in, '
                        'toggling zoom off instead')
            self.toggle_zoom()
            return
        idx = self._ctx.tui_state.selected_process_idx
        current_focus = self.get_focused_widget()
        if current_focus == FocusWidget.TERMINAL:
            self.focus_to_sidebar()
        elif current_focus == FocusWidget.SIDE_BAR_SELECT and idx is not None:
            self.focus_to_terminal()

    def focus_to_sidebar(self):
        logger.info('focusing on sidebar')
        side_bar = self.get_element_by_focus_widget(FocusWidget.SIDE_BAR_SELECT)
        if side_bar:
            application = get_app()
            application.layout.focus(side_bar)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read()
setup(
    name='procmux',
    version='1.0.6',
    author='Nick Pisani',
    author_email='napisani@yahoo.com',
    license='MIT',
    description='a TUI utility for running multiple commands in parallel in easily switchable terminals',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/napisani/procmux',
    py_modules=['procmux', 'app'],
    packages=find_packages(),
    install_requires=[requirements],
    include_package_data=True,  # MANIFEST.in controls inclusions and exclusions
    python_requires='>=3.7',
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    entry_points='''
        [console_scripts]
        procmux=procmux:start_cli
    '''
)



from typing import Callable, Optional

from prompt_toolkit.key_binding import KeyBindings

from app.context import ProcMuxContext
from app.log import logger
from app.tui.focus import FocusManager


def register_configured_keybinding(key_config_attr: str, handler: Callable,
                                   kb: Optional[KeyBindings] = None) -> KeyBindings:
    if not kb:
        kb = KeyBindings()
    ctx = ProcMuxContext()
    all_keys = getattr(ctx.config.keybinding, key_config_attr)
    if 'disabled' in all_keys:
        return kb

    for keybinding in all_keys:
        @kb.add(keybinding)
        def _toggle_scroll(event) -> None:
            handler(event)

    return kb


def register_app_wide_configured_keybindings(focus_manager: FocusManager,
                                             kb: Optional[KeyBindings] = None) -> KeyBindings:
    if not kb:
        kb = KeyBindings()
    ctx = ProcMuxContext()

    def _switch_focus(_event):
        logger.info('in _switch_focus')
        if not focus_manager.is_focused_on_free_form_input():
            focus_manager.toggle_sidebar_terminal_focus()

    kb = register_configured_keybinding('switch_focus', _switch_focus, kb)

    def _zoom(_event):
        logger.info('in _zoom')
        if not focus_manager.is_focused_on_free_form_input():
            focus_manager.toggle_zoom()

    kb = register_configured_keybinding('zoom', _zoom, kb)

    def _toggle_scroll(_event) -> None:
        logger.info('in _toggle_scroll')
        proc_idx = ctx.tui_state.selected_process_idx
        if not focus_manager.is_focused_on_free_form_input() and proc_idx > -1:
            scroll_mode_on = ctx.tui_state.terminal_managers[proc_idx].toggle_scroll_mode()
            if not scroll_mode_on:
                focus_manager.focus_to_sidebar()

    kb = register_configured_keybinding('toggle_scroll', _toggle_scroll, kb)
    return kb



import argparse

parser = argparse.ArgumentParser(description="procmux-tui")
parser.add_argument('--config',
                    required=True)
parser.add_argument('--config-override',
                    required=False)
cli_args = parser.parse_args()





import logging

from app.config import ProcMuxConfig
from app.context import ProcMuxContext
from app.log import formatter, logger


def run_app(cfg: ProcMuxConfig):
    if cfg.log_file:
        file_handler = logging.FileHandler(cfg.log_file)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    ProcMuxContext().bootstrap(cfg)
    from app.tui.app import start_tui
    start_tui()


